#include <iostream>
#include <assert.h>
#include "filenamedocparser.h"
#include "sentenceparser.h"
#include "citeseerparser.h"
#include "docnomanager.h"

unsigned DocParser::_output_offset = 0;

DocParser*
alloc_parser (string which, unsigned l, int i, FILE *logf, FILE *ffp, FILE *ffindp)
{
  if (which == "FilenameDocParser") 
    return new FilenameDocParser(l, i, logf, ffp, ffindp);
  else if (which == "DocParser") 
    return new DocParser(l, i, logf, ffp, ffindp);
  else if (which == "SentenceParser")
    return new SentenceParser(l, i, logf, ffp, ffindp);
  else if (which == "CiteseerParser")
     return new CiteseerParser(l, i, logf, ffp, ffindp);
  else
    return NULL;
}

char 
DocParser::which_parser(string fname)
{
  if (strcmp(fname.c_str(),"http://")) {
    return DOC_HTML;
  }else {
    return DOC_UNKNOWN;
  }
}


DocParser::DocParser (unsigned l, int i, FILE *f, FILE *ffp, FILE *ffindp)
  : _maxwordlen(l),_maxwordindexed(i), _logfp(f), _totalwords(0), _totalc(0), _last_dn(0) 
{
  _ffp = ffp;
  if (_ffp) {
    _f_offset = ftello (_ffp);
  } else
    _f_offset = 0;
  _ffindp = ffindp;
  _last_offset = _f_offset;
}

bool
DocParser::getword(word_pos &wc, char *word, FILE *fp, unsigned fsz)
{
  int c;
  unsigned wordi = 0;
  bool inword = false;
  bool last_is_space = false;

  while ((_curr_wc < _maxwordindexed) 
      && (!_output_offset || _totalc < _output_offset)
      && (!fsz || _totalc < fsz) 
      && ((c = getc(fp))!=EOF)) {

    _totalc++;

    if (_ffp) {
      if (isprint(c)) {
	 putc(c, _ffp);
	 last_is_space = (c == ' ');
	 _f_offset++;
	 _total_off++;
      } else {
	 if (!last_is_space) {
	    putc(' ', _ffp);
	    last_is_space = true;
	    _f_offset++;
	    _total_off++;
	 }
      }
    }

    if (WORDCHAR(c)) {
      c = tolower(c);
      if (wordi < _maxwordlen)
	word[wordi++] = (char)c;
      if (!inword) 
	inword = true;
    } else if (inword) {
      //output indexing information for this word
      word[wordi] = '\0';
      if (_output_offset) {
	 wc = (_total_off - wordi -1);
	 assert (_total_off + _last_offset == _f_offset);
      } else
	wc = _curr_wc;
      _curr_wc++;
      return true;
    }
  }
  return false;
}

unsigned 
DocParser::parse(string fname, FILE *fp, unsigned fsz, DID dd)
{
  /* use the c interface because it gives me offset information */
  _curr_wc = _totalc = _total_off = 0;
  char word[_maxwordlen+1]; 

  DID dc;
  if (!dd)
    dc = DocNoManager::Instance()->getDocNo(fname); 
  else
    dc = dd;

  if (!dc) {
    if (_logfp)
      fprintf(_logfp, "error: %s exists in DB\n", fname.c_str());
    return 0; //file exists in db
  }

  //write offset information to file
  if (_ffp) {
     //this is probably redundant if DIDs are always from 1..N
     assert(!_last_dn || dc == (_last_dn + 1));
     xwrite(&_f_offset, sizeof(_f_offset), _ffindp);
  }

  _last_dn = dc;

  word_pos wc;
  int wlen;
  while (wlen = getword(wc, word, fp, fsz)) {
    fputs(word, stdout);
    putc('\0', stdout);
    xwrite(&dc,sizeof(DID), stdout);
    xwrite(&wc,sizeof(word_pos),stdout);
    _totalwords++;
  }

  if (_logfp) {
    if (!_totalc)
      fprintf (_logfp, "error: empty file? %s\n", fname.c_str ());
    else {
       int allsz = fsz;
      if (!allsz)
	 allsz = lseek(fileno(fp), 0L, SEEK_END);
      fprintf (_logfp,"%s dc:%u wc:%u max:%u totalc:%u sz:%u eof:%d\n", 
	  fname.c_str (), dc, (unsigned)wc, _maxwordindexed, _totalc, allsz, feof(fp)?1:0);
    }
  }

  _last_offset = _f_offset;
  return _totalc;
}

//return sz of char scanned
unsigned
getline (char *buf, unsigned maxbuflen)
{
  char c;
  unsigned sz = 0;
  while (1) {
    c = getchar ();
    if ((c == EOF) || (c == '\n') || (sz == maxbuflen-1)) {
      buf[sz++] = '\0';
      return sz;
    }
    buf[sz++] = c;
  }
  return sz;
}

vector<string>
split (char *buf, char sep, char maxf)
{
  vector<string> v;
  v.clear ();
  unsigned i,s;
  i = s = 0;
  while (buf[i] != '\0') {
    if (buf[i] == sep) {
      buf[i] = '\0';
      if (i > s)
	v.push_back (string (buf + s));
      buf[i] = sep;
      s = i + 1;
    }else if (v.size() >= (maxf-1)) {
      v.push_back (string (buf + s));
      return v;
    }
    i++;
  }
  if (i > s)
    v.push_back ( string (buf + s));
  return v;
}


void
skipfile(FILE *fp, int sz)
{
  if (fp == stdin) {
    for (uint i = 0; i < sz; i++) 
      getc(stdin);
  }else{
    fseek(fp,sz,SEEK_CUR);
  }
}


int
main(int argc, char **argv)
{
  bool update_only = false;
  bool txtfrominput = false;

  char *config = "mkdb.config";
  for (unsigned i = 1; i < argc; i++) {
    if (strcmp(argv[i],"-f") == 0 && i+1 < argc) {
      i++;
      config = argv[i];
    }else if (strcmp(argv[i],"-u") == 0) {
      update_only = true;
      i++;
    }else if (strcmp(argv[i],"-i") == 0) {
      txtfrominput = true;
      i++;
    } else {
      cerr << "usage: docparser [-i txtfrominput] [-f configfile] [-u update-only]" << endl;
      exit(1);
    }
  }

  Args *a = new Args(config);

  string prefix = a->sget("prefix","ind");
  string no2name = prefix;
  no2name.append("-n2f.db");
  string name2no = prefix;
  name2no.append("-f2n.db");

  FILE *ffp = NULL;
  FILE *ffindp = NULL;

  unsigned maxwordlen  = a->nget<unsigned>("maxwordlen",200);
  unsigned maxwordind = a->nget<unsigned>("maxwordindexed",(word_pos)(-1));
  unsigned numtags = a->nget<unsigned>("numtags",0);
  if (numtags && maxwordind > (1<<(sizeof(word_pos)*8-numtags))) {
    maxwordind = 1<<(sizeof(word_pos)*8-numtags);
    cerr << numtags << " tags in word_pos " << 
      " maxwordlen truncated to " << maxwordind << endl;
  }

  int output_offset =  a->nget("outputoffset",0);

  if (output_offset) {
     int maxoff = 1<<(sizeof(word_pos)*8-numtags);
     maxoff--;
     cerr << " output file offset instead of word position up to " << maxoff << endl;
     DocParser::setoutputoffset(maxoff);

     string ffname = prefix;
     ffname.append("-ff");
     if (update_only) {
	ffp = fopen (ffname.c_str(), "w");
	fseek(ffp, 0L, SEEK_END);
     } else 
	ffp = fopen (ffname.c_str(),"w+");
     ffname = prefix;
     ffname.append("-ffi");
     if (update_only) {
	ffindp = fopen(ffname.c_str(), "w");
	fseek(ffindp, 0L, SEEK_END);
      } else
	 ffindp = fopen (ffname.c_str(), "w+");
  }

  string log = a->sget("logfileprefix","none");
  FILE *logfp;
  if (log == "none") {
    logfp = NULL;
  } else {
    string logname = log;
    logname.append(".parse");
    logfp = fopen(logname.c_str(),"w");
    assert(logfp);
  }
  
  DocNoManager *m = new DocNoManager(no2name, name2no, update_only); 
  if( m != NULL ) {
    cerr << "docparser: " << (update_only? "updating ":"creating ")
	 << " DID offset " << DocNoManager::Instance()->getmaxfo() << endl;
  }

  string which_parser = a->sget("parser",string("DocParser"));
  DocParser *doc_parser = alloc_parser (which_parser, maxwordlen, maxwordind, logfp, ffp, ffindp);
  if (!doc_parser) {
    cerr << "no parser" << endl;
    exit(1);
  }

  
#define MAXBUFSZ 600
#define SZBUF_SZ 20
  string preamble = "!@#$%";
  char buf[MAXBUFSZ];
  vector<string> fields;
  unsigned long offset = 0;
  unsigned sz = 0;
  unsigned long totalf = 0;
  unsigned parsed = 0;
  FILE *fp = NULL;

  while (!feof(stdin)) {
    sz = getline (buf, MAXBUFSZ);
    assert (sz < MAXBUFSZ);
    offset += sz;
    if (sz <= 1) 
      continue;
    if (txtfrominput) {
      fields = split (buf, ' ', 3);
      assert (fields.size() >= 3);
      if (fields[0] != "!@#$%") {
	cerr << "doc out of sync at offset " << offset << " wrong preamble "
	  << fields[0] << endl;
	delete doc_parser; 
	exit (1);
      }
      unsigned fsz = atoi (fields[1].c_str ());
      if (fsz > 0) {
	parsed = doc_parser->parse (fields[2], stdin, fsz);
	if (parsed < fsz) 
	  skipfile (stdin, fsz-parsed);
	if (parsed > 0)
	  totalf ++;
      }
    } else {
      fields = split (buf, ' ', 2);
      if (fields.size () >= 1) {
	fp = fopen (fields[0].c_str (), "r");
	if (!fp) {
	  if (logfp)
	    fprintf (logfp, " cannot open file %s\n", fields[0].c_str ());
	} else {
	  parsed = doc_parser->parse (string (buf), fp, 0);
	  if (parsed > 0)
	    totalf++;
	  fclose (fp);
	}
      }
    }
  }
  cerr << " maxDID " << DocNoManager::Instance()->getmaxfo() 
    << " total files parsed " << totalf << " words " << doc_parser->totalwords() 
    << " sizeof(ind_offset) " << sizeof(ind_offset)
    << endl;

DONE:
  if (doc_parser)
    delete doc_parser;
  if (m)
    delete m;
  delete a;

  if (logfp)
     fclose(logfp);
  if (ffp)
     fclose (ffp);
  if (ffindp)
     fclose (ffindp);
}
