/* Robert Morris (modified by Jinyang)
 * 
 * This sort relies on the assumption that input comes in already
 * sorted by: document id (DID) and position within document
 *
 * write the inverted file into a regular file
 * and output <word,offset> to stdout
 *
 * original note --------------
 *
 * Like nisort, but use a hash table for the first in-core
 * phase of the sort. Then merge-sort the resulting temporary
 * files. The point is that this method takes advantage of
 * the fact that the original input already comes sorted
 * by position. And the fact that there's no particular point
 * in maintaining alphabetical order -- just some order.
 * We still have to qsort() the words (ie hash buckets),
 * but there are far fewer distinct words than total words.
 * The first phase runs about twice as fast as nisort.
 * 
 * Idea by Eddie Kohler.
 * 
 * Could avoid the sort() if each bucket were a linked list
 * of the words that hashed to that value, and keep the list
 * in sort order.
 *
 * When merging, once all files (or a subset?) are on the
 * same word, just consider the positions without redundant
 * strcmp()s. Merge now three times as fast!
 *
 * Eddie points out that all the positions in temporary file
 * K are less than in K+1, so no true merge is required for
 * the positions.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include "args.h"
#include "mkdb.h"

char *tmpdir;
char outfile[100];
char oldoutfile[100];
struct timeval tv0;
extern int errno;
unsigned maxwordlen;

bool update_only;

void cleanup(int exit_status);
void pass0();
void passN();
void printrusage(int incremental);

int
main(int argc, char *argv[])
{
  int i;
  char ch;

  gettimeofday(&tv0, (struct timezone *) 0);

  tmpdir = "/tmp";
  if(getenv("TMPDIR"))
    tmpdir = getenv("TMPDIR");

  char *config = "mkdb.config";
  update_only = false;

  while ((ch = getopt(argc, argv, "t:f:u")) != -1) {
    switch (ch) {
      case 't':
	tmpdir = optarg;
	break;
      case 'f':
	config = optarg;
	break;
      case 'u':
	update_only = true;
	break;
      default:
	fprintf(stderr,"./edsort [-t tmpdir] [-u (update)] [-f config_file]\n");
	exit(1);
    }
  }

  Args *a = new Args(config);
  maxwordlen = a->nget<unsigned>("maxwordlen", 100);
  string prefix = (a->sget("prefix","ind")).c_str();
  strcpy(outfile,prefix.c_str());
  strcpy(outfile+prefix.size(), "-f");
  strcpy(oldoutfile, outfile);
  strcpy(outfile+strlen(outfile),".tmp");

  signal(SIGINT, cleanup);
  signal(SIGTERM, cleanup);
  signal(SIGPIPE, SIG_IGN);

  pass0();
  passN();
  
  fprintf(stderr, "edsort: total: ");
  printrusage(0);
  fprintf(stderr, "\n");

  exit(0);
}

/*
 * OpenBSD fwrite() is pretty slow for short writes.
 */
int
xwrite(const void *xptr, int n, FILE *fp)
{
  int i;
  const char *ptr = (const char *) xptr;

  for(i = 0; i < n; i++)
    if(putc(ptr[i], fp) == EOF)
      return(EOF);
  return(i);
}

int pass0files = 0;

void
sorttmpname(int pass, int fn, char *buf)
{
  sprintf(buf, "%s/nis-%d-%d-%d", tmpdir, getpid(), pass, fn);
}

/*
 * First pass (in-core sort):
 */

#define BLOCKSIZE 128
struct Block {
  struct Block *next;
  int n; //number of groups of Tags
  PostIt* p[BLOCKSIZE];
};

struct Bucket {
  char *word;
  struct Block *b0;
  struct Block *bN;
  unsigned n; //number of blocks
};

#define HASHSIZE (1024*1024) //1M of hash table entries
struct Bucket *table;

int
lookup(char *word)
{
  int i;
  unsigned int h = 0;

  for(i = 0; word[i]; i++)
    h = (h * 33) + word[i];
  h &= (HASHSIZE-1);
  for(i = 0; i < HASHSIZE; i++){
    if(table[h].word == 0)
      return(h);
    if(strcmp(table[h].word, word) == 0)
      return(h);
    h++;
    if(h >= HASHSIZE)
      h = 0;
  }
  fprintf(stderr, "edsort: hash table full\n");
  exit(1);
}

int
tablecompar(const void *xa, const void *xb)
{
  struct Bucket *a = (struct Bucket *) xa;
  struct Bucket *b = (struct Bucket *) xb;

  if(a->word && b->word)
    return(strcmp(a->word, b->word));
  if(a->word == 0)
    return(1);
  if(b->word == 0)
    return(-1);
  return(0);
}

void
flushwords()
{
  int i, pi;
  struct Block *bl;
  char tn[512];
  FILE *fp;

  sorttmpname(0, pass0files++, tn);
  fp = fopen(tn, "w");
  if(fp == 0){
    fprintf(stderr, "ni: cannot write %s\n", tn);
    cleanup(1);
  }

#ifdef HAVE_MERGESORT
  mergesort(table, HASHSIZE, sizeof(table[0]), tablecompar);
#else
  qsort(table, HASHSIZE, sizeof(table[0]), tablecompar);
# endif // HAVE_MERGESORT

  for(i = 0; i < HASHSIZE; i++){
    struct Bucket *bu = &table[i];
    if(bu->word == 0)
      break;

    xwrite(bu->word, strlen(bu->word) + 1, fp);
    xwrite(&(bu->n), sizeof(bu->n), fp);

    for(bl = bu->b0; bl; bl = bl->next){
      for(pi = 0 ; pi < bl->n; pi++){
        if(xwrite(bl->p[pi], sizeof(PostIt), fp) <= 0){
          fprintf(stderr, "edsort: xwrite failed\n");
          cleanup(1);
        }
      }
    }

    bu->word = 0;
  }

  if(fclose(fp) != 0){
    fprintf(stderr, "edsort: fclose failed\n");
    cleanup(1);
  }
}
/*
#define MAXWORDBYTES (1 << 20) //1M
#define MAXINFO (1 << 20) //1M
#define MAXBLOCKS (1 << 20) //1M 
*/
#define MAXWORDBYTES (1 << 27) //128M
#define MAXINFO (1 << 27) //128M
#define MAXBLOCKS (1 << 27) //128M 

void
pass0()
{
  int maxinfo;
  char *wordbytes,*p;
  int wordi, infoi;
  int maxblocks;
  struct Block *blocks;
  int blocki;
  int buckets_used = 0;
  struct Bucket *bu;
  struct Block *bl;
  PostIt *infobuf, *infop;
  int len, c;
  unsigned h,xx;
  unsigned wc = 0;

  maxinfo = MAXINFO/sizeof(PostIt); //32 * sizeof(PostIt) = 192M

  wordbytes = (char *)malloc(MAXWORDBYTES);
  infobuf = (PostIt *)malloc(sizeof(PostIt)*maxinfo);
  table = (struct Bucket *)malloc(sizeof(struct Bucket) * HASHSIZE);

  wordi = 0;
  infoi = 0;

  maxblocks = MAXBLOCKS/ (sizeof(struct Block)); 
  blocks = (struct Block *) malloc(maxblocks * sizeof(struct Block));
  blocki = 0;

  assert(wordbytes && infobuf && blocks && table);

  while(1){
    if((wordi + maxwordlen + 1) > MAXWORDBYTES||
       (infoi + 1) > maxinfo ||
       blocki + 1 >= maxblocks ||
       buckets_used >= HASHSIZE) {
      flushwords();
      wordi = 0;
      infoi = 0;
      blocki = 0;
      buckets_used = 0;
    }

    len = 0;
    p = wordbytes + wordi;
    while((c = getchar()) != EOF) {
      p[len++] = c;
      if (c == '\0') {
	break;
      }
    }
    wc++;
    if(len == 0) {
      assert(c == EOF);
      break;
    }
    assert(len>1);

    infop = infobuf + infoi;
    if (fread(infop, sizeof(PostIt),1, stdin)!=1) {
      fprintf(stderr,"reading error\n");
      exit(1);
    }
    infoi++;
      
    h = lookup(p);
    bu = &table[h];
    
    if(bu->word == 0){
      buckets_used++;
      bu->word = p;
      wordi += len;
      bu->n = 0;
      bl = bu->b0 = bu->bN = &blocks[blocki++];
      bl->n = 0;
      bl->next = 0;
    } else if(bu->bN->n >= BLOCKSIZE){
      bl = bu->bN->next = &blocks[blocki++];
      bu->bN = bl;
      bl->next = 0;
      bl->n = 0;
    } else {
      bl = bu->bN;
    }

    bl->p[bl->n] = infop;
    bl->n += 1;
    bu->n += 1;

  }

  flushwords();

  fprintf(stderr, "edsort: pass 0: bucket_used %u: ", buckets_used);
  printrusage(1);
  fprintf(stderr, "\n");

  delete wordbytes;
  delete infobuf;
  delete blocks;
  delete table;
  table = NULL;
}

/*
 * Read a '\0'-terminated word.
 * Return 0 on success, EOF on EOF.
 */
int
zword(FILE *fp, char buf[], int bufsize)
{
  int i, c;

  i = 0;
  while((c = getc(fp)) != EOF){
    if(i < bufsize)
      buf[i++] = c;
    if(c == '\0')
      break;
  }
  if (i == bufsize)
    buf[bufsize-1] = '\0';
  if(i > 0)
    return(0);
  return(EOF);
}

/*
 * Read an unsigned int
 */
inline unsigned
znum(FILE *fp)
{
  unsigned pos;
  fread(&pos, sizeof(pos), 1, fp);
  return(pos);
}

/*
 * Copy bytes from ifp to ofp.
 */
void
zcopy(FILE *ifp, FILE *ofp, unsigned n)
{
  int c;
  unsigned i;

  for(i = 0; i < n && (c = getc(ifp)) != EOF; i++){
    if(putc(c, ofp) == EOF){
      fprintf(stderr, "edsort: putc: %s\n", strerror(errno));
      cleanup(1);
    }
  }
}

void
passN()
{
  int opass = 0;
  int onf = pass0files;
  int nnf = 0;
  FILE *ofp;
#define MaxFDS 800
  FILE *fps[MaxFDS];
  int ifn;
  char iname[256], oname[256];
  char word[MaxFDS][maxwordlen+1];
  int eof[MaxFDS];
  unsigned numwords = 0;
  bool done = false;
  ind_offset offset = 0;
  char *minmask = NULL;

  while(1){
    nnf = 0;
    ifn = 0;
    while(ifn < onf){
      int xni, fpn, startfpn;

      if(ifn == 0 && (ifn + MaxFDS + 1) >= onf) {
        strcpy(oname, outfile);
        ofp = fopen(outfile,"w");
	done = true;
	if (!ofp) {
	  cleanup(1);
	  exit(1);
	}
        nnf++;
      } else {
        sorttmpname(opass + 1, nnf++, oname);
        ofp = fopen(oname, "w");
        if(ofp == 0){
          fprintf(stderr, "edsort: cannot write %s\n", oname);
          cleanup(1);
        }
      }

      xni = onf - ifn;
      if(xni > MaxFDS)
        xni = MaxFDS;

      if (update_only && done) {
	fps[0] = fopen(oldoutfile, "r");
	if (fps[0] == 0) {
	  fprintf(stderr,"edsort: WARNING cannot open old index %s\n", oldoutfile);
	  update_only = false;
	}else {
	  eof[0] = zword(fps[0], word[0], sizeof(word[0]));
	}
	startfpn = 1;
	xni++;
      }else
	startfpn = 0;

      for (fpn = startfpn; fpn < xni; fpn++) {
        sorttmpname(opass, ifn + fpn - startfpn, iname);
        fps[fpn] = fopen(iname, "r");
        if(fps[fpn] == 0){
          fprintf(stderr, "edsort: cannot open %s\n", iname);
          cleanup(1);
        }
        eof[fpn] = zword(fps[fpn], word[fpn], sizeof(word[fpn]));
      }

      /*
       * Merge sort on the MaxFDS inputs.
       */
      while(1){
        /*
         * Find the minimum word, and find all the temporaries
         * with that word. We need them all to pre-compute the
         * total number of positions.
         */
        unsigned n[MaxFDS], nsum;
        const char *minword = 0;

        minmask = (char *)realloc(minmask,sizeof(char)*xni);
	assert(minmask);
	bzero(minmask, sizeof(char)*xni);

        for(fpn = 0; fpn < xni; fpn++){
          if(eof[fpn] == 0){
            if(minword == 0){
              minword = word[fpn];
              minmask[fpn] = 1;
            } else {
              int x = strcmp(word[fpn], minword);
              if(x < 0){
                minword = word[fpn];
		bzero(minmask,sizeof(char)*xni);
                minmask[fpn] = 1;
              } else if(x == 0){
                minmask[fpn] = 1;
              }
            }
          }
        }

        if(minword == 0)
          break;

        nsum = 0;
        for(fpn = 0; fpn < xni; fpn++){
          if(minmask[fpn] == 1) {
            n[fpn] = znum(fps[fpn]);
            nsum += n[fpn];
          }
        }

        fputs(minword, ofp);
        putc('\0', ofp);
        xwrite(&nsum, sizeof(nsum), ofp);

	if (done) {
	  numwords++;
	  fputs(minword,stdout);
	  putc('\0',stdout);
	  xwrite(&offset,sizeof(ind_offset), stdout);
	  offset += strlen(minword) + 1 + sizeof(nsum) + nsum * sizeof(PostIt);
	}

        /*
         * Spit out the position lists, first temporary first.
         * Advance to the next word.
         */
        for(fpn = 0; fpn < xni; fpn++){
          if(minmask[fpn]==1) {
            unsigned bytes = n[fpn] * sizeof(PostIt);
            zcopy(fps[fpn], ofp, bytes);
            eof[fpn] = zword(fps[fpn], word[fpn], sizeof(word[fpn]));
          }
        }
      }

      for(fpn = 0; fpn < xni; fpn++){
        fclose(fps[fpn]);
	if (startfpn && fpn == 0)
	  continue;
        sorttmpname(opass, ifn + fpn - startfpn, iname);
        if(unlink(iname) != 0){
          fprintf(stderr, "edsort: unlink %s failed\n", iname);
          cleanup(1);
        }
      }
      ifn += MaxFDS;
      
      if(fclose(ofp) != 0){
        fprintf(stderr, "edsort: write to %s failed\n", oname);
        cleanup(1);
      }
      if(done){
        fprintf(stderr, "edsort: merged %d files, %u words, ", onf, numwords);
        printrusage(1);
        fprintf(stderr, "\n");
	rename(outfile, oldoutfile);
        return;
      }
    }

    fprintf(stderr, "edsort: merged %d files: ", onf);
    printrusage(1);
    fprintf(stderr, "\n");

    opass++;
    onf = nnf;
  }
}

void
cleanup(int exit_status)
{
  int pass, passfiles, i;

  pass = 0;
  passfiles = pass0files;

  while(pass == 0 || passfiles > 1){
    for(i = 0; i < passfiles; i++){
      char name[128];
      sorttmpname(pass, i, name);
      unlink(name);
    }
    passfiles = (passfiles + MaxFDS - 1) / MaxFDS;
    pass++;
  }
  unlink(outfile);
  exit(exit_status);
}

void
printrusage(int incremental)
{
  static struct rusage oru;
  static double _or = 0.0;
  struct rusage ru;
  struct timeval tv;
  double u, s, r;
  long i, o;

  gettimeofday(&tv, (struct timezone *) 0);
  if(getrusage(RUSAGE_SELF, &ru) < 0){
    perror("edsort: getrusage");
    return;
  }

  u = ru.ru_utime.tv_sec + (ru.ru_utime.tv_usec / 1000000.0);
  s = ru.ru_stime.tv_sec + (ru.ru_stime.tv_usec / 1000000.0);
  r = (tv.tv_sec + tv.tv_usec / 1000000.0) -
    (tv0.tv_sec + tv0.tv_usec / 1000000.0);
  i = ru.ru_inblock;
  o = ru.ru_oublock;

  if(incremental){
    u -= oru.ru_utime.tv_sec + (oru.ru_utime.tv_usec / 1000000.0);
    s -= oru.ru_stime.tv_sec + (oru.ru_stime.tv_usec / 1000000.0);
    r -= _or;
    i -= oru.ru_inblock;
    o -= oru.ru_oublock;
    oru = ru;
    _or += r;
  }

  fprintf(stderr, "%.2fu %.2fs %.2fr %ldi %ldo", u, s, r, i, o);
}
